#define F_CPU 16000000
#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <util/atomic.h>
#include <time.h>

#define A PC0
#define B PC1
#define C PC2
#define D PC3
#define E PC4
#define F PC5
#define G PC6
#define DIGIT1 PD6
#define DIGIT2 PD5
#define DIGIT3 PD4
#define DIGIT4 PD1
#define SER PB0
#define SRCLK PB1
#define SRCLR PB3
#define BUZZER_PORT PORTB
#define BUZZER PB4
#define BUZZER_DDR DDRB
#define JOYSTICK_BUTTON PD0

int led_matrix[8][8], x = 512, y = 512, snk_head_x = -1, snk_head_y = -1, food_x, food_y;
volatile int snake_segments = 3, c = 1;

struct led {
    int x;
    int y;
    int old_x;
    int old_y;
    int direction; //0 up / 1 down / 2 left / 3 right
    led *parent;
    led *child;
};

struct Food {
    int x;
    int y;
};

struct adc_res {
    int value_x;
    int value_y;
};

void blink() {
    PORTD |= (1 << PD7);
    _delay_ms(500);
    PORTD &= ~(1 << PD7);
    _delay_ms(500);
}


void reset_7_segm() {
    PORTC |= (1 << A); 
    PORTC |= (1 << B); 
    PORTC |= (1 << C); 
    PORTC |= (1 << D); 
    PORTC |= (1 << E); 
    PORTC |= (1 << F); 
    PORTC |= (1 << G); 
}

void zero() {
    reset_7_segm();
    PORTC &= ~(1 << A); 
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << C); 
    PORTC &= ~(1 << D); 
    PORTC &= ~(1 << E); 
    PORTC &= ~(1 << F); 
}

void one() {
    reset_7_segm();
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << C); 
}

void two() {
    reset_7_segm();
    PORTC &= ~(1 << A); 
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << D); 
    PORTC &= ~(1 << E);
    PORTC &= ~(1 << G);    
}

void three() {
    reset_7_segm();
    PORTC &= ~(1 << A); 
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << C); 
    PORTC &= ~(1 << D); 
    PORTC &= ~(1 << G); 
}

void four() {
    reset_7_segm();
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << C); 
    PORTC &= ~(1 << F); 
    PORTC &= ~(1 << G); 
}

void five() {
    reset_7_segm();
    PORTC &= ~(1 << A); 
    PORTC &= ~(1 << C); 
    PORTC &= ~(1 << D); 
    PORTC &= ~(1 << F); 
    PORTC &= ~(1 << G); 
}

void six() {
    reset_7_segm();
    PORTC &= ~(1 << A);
    PORTC &= ~(1 << C); 
    PORTC &= ~(1 << D); 
    PORTC &= ~(1 << E); 
    PORTC &= ~(1 << F); 
    PORTC &= ~(1 << G); 
}

void seven() {
    reset_7_segm();
    PORTC &= ~(1 << A); 
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << C); 
}

void eight() {
    reset_7_segm();
    PORTC &= ~(1 << A); 
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << C); 
    PORTC &= ~(1 << D); 
    PORTC &= ~(1 << E); 
    PORTC &= ~(1 << F); 
    PORTC &= ~(1 << G); 
}

void nine() {
    reset_7_segm();
    PORTC &= ~(1 << A); 
    PORTC &= ~(1 << B); 
    PORTC &= ~(1 << C); 
    PORTC &= ~(1 << D); 
    PORTC &= ~(1 << F); 
    PORTC &= ~(1 << G); 
}

void init_adc() {
    //set internal reference 2.56V
    ADMUX &= ~(1 << REFS1);
    ADMUX |= (1 << REFS0);

    //right allignment
    ADMUX &= ~(1 << ADLAR);

    //ADC enable
    ADCSRA |= (1 << ADEN);

    //Prescaler 128
    ADCSRA |= (1 << ADPS2);
    ADCSRA |= (1 << ADPS1);
    ADCSRA |= (1 << ADPS0);

    //Interrupt enable
    ADCSRA |= (1 << ADIE);
}

uint16_t adc_convert(int adc) {
    if (adc == 0) {
        //ADC0
        ADMUX &= ~(1 << MUX4);
        ADMUX &= ~(1 << MUX3);
        ADMUX &= ~(1 << MUX2);
        ADMUX &= ~(1 << MUX1);
        ADMUX &= ~(1 << MUX0);
    } else if (adc == 1) {
        //ADC1
        ADMUX &= ~(1 << MUX4);
        ADMUX &= ~(1 << MUX3);
        ADMUX &= ~(1 << MUX2);
        ADMUX &= ~(1 << MUX1);
        ADMUX |= (1 << MUX0);
    }

    ADCSRA |= (1 << ADSC);         //start conversion
    loop_until_bit_is_set(ADCSRA, ADIF);  //wait until conversion is done: until ADIF from ADCSRA is set to 1
    uint16_t value = ADC;

    return value;
}

void init_7_segm() {
    //7 segmente
    //DDRD |= (1 << DIGIT1);
    DDRD |= (1 << DIGIT2);
    DDRD |= (1 << DIGIT3);
    //DDRD |= (1 << DIGIT4);

    DDRC |= (1 << A);
    DDRC |= (1 << B);
    DDRC |= (1 << C);
    DDRC |= (1 << D);
    DDRC |= (1 << E);
    DDRC |= (1 << F);
    DDRC |= (1 << G);

    //PORTD |= (1 << DIGIT1);
    //PORTD |= (1 << DIGIT2);
    //PORTD |= (1 << DIGIT3);
    //PORTD |= (1 << DIGIT4);
}

void show_digit(int digit) {
    if (digit == 0) {
        zero();
    }
    if (digit == 1) {
        one();
    }
    if (digit == 2) {
        two();
    }
    if (digit == 3) {
        three();
    }
    if (digit == 4) {
        four();
    }
    if (digit == 5) {
        five();
    }
    if (digit == 6) {
        six();
    }
    if (digit == 7) {
        seven();
    }
    if (digit == 8) {
        eight();
    }
    if (digit == 9) {
        nine();
    }
}

void shiftDataIn(uint8_t data) {
    if (data == 0) {
        PORTB &= ~(1 << SER);
        PORTB |= (1 << SRCLK);
        PORTB &= ~(1 << SRCLK);
    } else {
        PORTB |= (1 << SER);
        PORTB |= (1 << SRCLK);
        PORTB &= ~(1 << SRCLK);
    }
}

void copyShiftToStorage() {
    PORTB |= (1 << SRCLR);
    //_delay_loop_1(1);
    PORTB &= ~(1 << SRCLR);
    //_delay_loop_1(1);
}

void turn_on_led(int x, int y) {
    int i;
    uint8_t one = 1, zero = 0;

    for (i = 7; i >= 0; i--) {
        if (i == y) {
            shiftDataIn(one);
            copyShiftToStorage();
        } else {
            shiftDataIn(zero);
            copyShiftToStorage();            
        }
    }

    for (i = 7; i >= 0; i--) {
        if (i == x) {
            shiftDataIn(one);
            copyShiftToStorage();
        } else {
            shiftDataIn(zero);
            copyShiftToStorage();            
        }
    }
}

void reset_shift() {
    int i;
    uint8_t zero = 0;

    for (i = 14; i >= 0; i--) {
        shiftDataIn(zero);
        copyShiftToStorage();  
    }
}

void custom_delay(int delay, struct adc_res *adc_val) {
    int i;
    for (i = 0; i < delay; i++) {
        adc_val->value_x = adc_convert(0);
        adc_val->value_y = adc_convert(1);
        _delay_ms(1);
    }
}

void reset_matrix() {
    int i, j;

    for (i = 0; i < 8; ++i)
    {
        for (j = 0; j < 8; ++j) {
            led_matrix[i][j] = 0;
        }
    }
}

void insert_to_matrix(int x, int y) {
    led_matrix[x][y] = 1;
}

void show_number(int number) {
    if (number < 10) {
        PORTD &= ~(1 << DIGIT3);
        PORTD |= (1 << DIGIT2);
        show_digit(number);
        _delay_us(1000);
    } else {
        PORTD &= ~(1 << DIGIT3);
        PORTD |= (1 << DIGIT2);
        show_digit(number % 10);

        number = number / 10;

        _delay_us(500);
        PORTD &= ~(1 << DIGIT2);
        PORTD |= (1 << DIGIT3);
        show_digit(number % 10);
        _delay_us(500);
    }
}

void new_delay_ms(int count) {
  while(count--) {
    _delay_ms(1);

  }
}

void draw_matrix() {
    int i, j, k, led_refresh_rate = 1, delay = 500, ms = 0;

    for (k = 0; k < delay / led_refresh_rate / snake_segments; ++k)
    {
        cli();
        ADCSRA |= (1 << ADEN);
        for (i = 7; i >= 0; i--)
        {
            for (j = 7; j >= 0; j--)
            {
                ms++;
                if (snk_head_x == i && snk_head_y == j) {
                    turn_on_led(i, j);
                    _delay_ms(2);
                }

                if (led_matrix[i][j] == 1) {
                    reset_shift();
                    turn_on_led(i, j);
                    _delay_us(400); 
                    x = adc_convert(0);
                    y = adc_convert(1);                 
                    reset_shift();
                }
            }
        }
        ADCSRA &= ~(1 << ADEN);
        sei();
    }

    if (delay > ms) {
        new_delay_ms(1000 - ms);
    }
}

void generate_food(struct Food *food, struct led *snk_led) {
    int sem = 0;
    struct led *snk_cpy = snk_led;

    while (sem == 0) {
        food->x = rand() % 8;
        food->y = rand() % 8;

        snk_led = snk_cpy;
        sem = 1;
        while (snk_led != NULL) {
            if (snk_led->x == food->x && snk_led->y == food->y) {
                sem = 0;
                break;
            }
            snk_led = snk_led->child;
        }
    }
    food_x = food->x;
    food_y = food->y;
}

int isDead(struct led *snk_led) {
    int head_x = snk_led->x, head_y = snk_led->y;
    snk_led = snk_led->child->child;
    while (snk_led != NULL) {
        if (head_x == snk_led->x && head_y == snk_led->y)
            return 1;
        snk_led = snk_led->child;
    }
    return 0;
}

void init_buzzer() {
    BUZZER_DDR |= (1 << BUZZER);
}

void play_note() {
    int i;

    for (i = 0; i < 100; i++) {
        BUZZER_PORT |= (1 << BUZZER);
        _delay_us(500);
        BUZZER_PORT &= ~(1 << BUZZER);
        _delay_us(500);
    }
}

void turn_matrix_on() {
    int i;
    uint8_t one = 1;

    for (i = 15; i >= 0; i--) {
            shiftDataIn(one);
            copyShiftToStorage();
    }    
}

void turn_matrix_off() {
    int i;
    uint8_t zero = 0;

    for (i = 15; i >= 0; i--) {
            shiftDataIn(zero);
            copyShiftToStorage();
    }    
}

void deadAnimation() {
    int  score1 = (snake_segments - 3) / 10, score2 = (snake_segments - 3) % 10;

    c *= -1;
    turn_matrix_on();   
    _delay_ms(500);
    PORTD &= ~(1 << DIGIT2);
    PORTD |= (1 << DIGIT3);
    show_digit(score1);
    play_note();
    turn_matrix_off();
    _delay_ms(500);
    PORTD &= ~(1 << DIGIT3);
    PORTD |= (1 << DIGIT2);
    show_digit(score2);
}

void snake_open_mouth() {
    int i, k;
    uint8_t one = 1, zero = 0;

    reset_matrix();
    for (k = 0; k < 50; k++) {
        for (i = 7; i >= 0; i--) {
            shiftDataIn(one);
            copyShiftToStorage();
        }

        for (i = 6; i >= 0; i--) {
            shiftDataIn(zero);
            copyShiftToStorage();            
        }
        shiftDataIn(one);
        copyShiftToStorage();
        _delay_ms(1);

        reset_shift();
        for (i = 2; i >= 0; i--) {
            shiftDataIn(zero);
            copyShiftToStorage();
        }
        for (i = 4; i >= 0; i--) {
            shiftDataIn(one);
            copyShiftToStorage();
        }

        shiftDataIn(one);
        copyShiftToStorage();
        for (i = 6; i >= 0; i--) {
            shiftDataIn(zero);
            copyShiftToStorage();            
        }
        _delay_ms(1);

        turn_on_led(1, 0);
        _delay_ms(1);
        turn_on_led(2, 0);
        _delay_ms(1);
        turn_on_led(3, 0);
        _delay_ms(1);
        turn_on_led(4, 0);
        _delay_ms(1);
        turn_on_led(5, 0);
        _delay_ms(1);
        turn_on_led(6, 0);
        _delay_ms(1);


        turn_on_led(6, 4);
        _delay_ms(1);
        turn_on_led(5, 4);
        _delay_ms(1);
        turn_on_led(5, 3);
        _delay_ms(1);
        turn_on_led(2, 4);
        _delay_ms(1);
        turn_on_led(4, 3);
        _delay_ms(1);
        turn_on_led(4, 4);
        _delay_ms(1);
        turn_on_led(3, 3);
        _delay_ms(1);
        turn_on_led(3, 4);
        _delay_ms(1);
        turn_on_led(3, 6);
        _delay_ms(1);
        turn_on_led(4, 6);
        _delay_ms(1);
        turn_on_led(5, 5);
        _delay_ms(1);
    }

    draw_matrix();
}

void snake_close_mouth() {
   int i, k;
    uint8_t one = 1, zero = 0;

    reset_matrix();

    for (k = 0; k < 50; k++) {
        for (i = 7; i >= 0; i--) {
            shiftDataIn(one);
            copyShiftToStorage();
        }

        for (i = 6; i >= 0; i--) {
            shiftDataIn(zero);
            copyShiftToStorage();            
        }
        shiftDataIn(one);
        copyShiftToStorage();
        _delay_ms(1);

        reset_shift();
        for (i = 2; i >= 0; i--) {
            shiftDataIn(zero);
            copyShiftToStorage();
        }
        for (i = 4; i >= 0; i--) {
            shiftDataIn(one);
            copyShiftToStorage();
        }

        shiftDataIn(one);
        copyShiftToStorage();
        for (i = 6; i >= 0; i--) {
            shiftDataIn(zero);
            copyShiftToStorage();            
        }
        _delay_ms(1);

        turn_on_led(1, 0);
        _delay_ms(1);
        turn_on_led(2, 0);
        _delay_ms(1);
        turn_on_led(3, 0);
        _delay_ms(1);
        turn_on_led(4, 0);
        _delay_ms(1);
        turn_on_led(5, 0);
        _delay_ms(1);
        turn_on_led(6, 0);
        _delay_ms(1);


        turn_on_led(6, 4);
        _delay_ms(1);
        turn_on_led(5, 4);
        _delay_ms(1);
        turn_on_led(5, 3);
        _delay_ms(1);
        turn_on_led(2, 4);
        _delay_ms(1);
        turn_on_led(4, 3);
        _delay_ms(1);
        turn_on_led(4, 4);
        _delay_ms(1);
        turn_on_led(3, 3);
        _delay_ms(1);
        turn_on_led(3, 4);
        _delay_ms(1);
        turn_on_led(3, 5);
        _delay_ms(1);
        turn_on_led(4, 5);
        _delay_ms(1);
        turn_on_led(5, 5);
        _delay_ms(1);
    }
}

void start_game_animation() {
    snake_segments = 17;
    snake_close_mouth();
    snake_open_mouth();
    snake_segments = 3;
}

void move_snake(struct led *snk_led, struct adc_res *adc_val, struct Food *food) {
    struct led *snk_head = snk_led, *snk_tail;
    int tail_x, tail_y;

    snk_head_x = snk_head->x;
    snk_head_y = snk_head->y;
    reset_matrix();
    insert_to_matrix(food->x, food->y);

    while(snk_led != NULL)
    {
        insert_to_matrix(snk_led->x, snk_led->y);

        snk_led->old_x = snk_led->x;
        snk_led->old_y = snk_led->y;

        if (snk_led->parent == NULL) {
            if (snk_led->direction == 0) {
                snk_led->y--;
                if (snk_led->y == -1) {
                    snk_led->y = 7;
                }
            }

            if (snk_led->direction == 1) {
                snk_led->y++;
                if (snk_led->y == 8) {
                    snk_led->y = 0;
                }
            }

            if (snk_led->direction == 2) {
                snk_led->x--;
                if (snk_led->x == -1) {
                    snk_led->x = 7;
                }
            }

            if (snk_led->direction == 3) {
                snk_led->x++;
                if (snk_led->x == 8) {
                    snk_led->x = 0;
                }
            }
        } else {
            if (snk_led->child == NULL) {
                snk_tail = snk_led;
                tail_x = snk_led->x;
                tail_y = snk_led->y;
            }
            
            snk_led->x = snk_led->parent->old_x;
            snk_led->y = snk_led->parent->old_y;
        }

        snk_led = snk_led->child;
    }
    draw_matrix();
    if (snk_head->x == food->x && snk_head->y == food->y) {
        struct led *new_snk_led = (led*) malloc(sizeof(struct led));
        new_snk_led->x = tail_x;
        new_snk_led->y = tail_y;
        new_snk_led->direction = snk_tail->direction;
        new_snk_led->parent = snk_tail;
        new_snk_led->child = NULL;
        snk_tail->child = new_snk_led;

        snake_segments++;

        generate_food(food, snk_head);
        play_note();
    }
}

ISR (TIMER1_OVF_vect) // timer1 overflow interrupt
{
    int score1 = (snake_segments - 3) / 10, score2 = (snake_segments - 3) % 10;
    c *= -1;
    if (c == 1) {
        PORTD &= ~(1 << DIGIT2);
        PORTD |= (1 << DIGIT3);
        show_digit(score1);
    } else {
        PORTD &= ~(1 << DIGIT3);
        PORTD |= (1 << DIGIT2);
        show_digit(score2);
    }
}                    


void init_interrupt() {
    sei(); 
    TIMSK1 |= (1 << TOIE1);
    //enable interrupts
    TCCR1B |= (1 << CS10);
    // set prescaler to 64 and start the timer
}

int is_move_right(struct led *snk_led, int direction) {
    int new_x = snk_led->x, new_y = snk_led->y;
    if ((direction == 0 && snk_led->direction == 1) || (direction == 1 && snk_led->direction == 0) ||
        (direction == 2 && snk_led->direction == 3) || (direction == 3 && snk_led->direction == 2))
        return 0;

    if (direction == 0) {
        new_y--;
    }
    if (direction == 1) {
        new_y++;
    }
    if (direction == 2) {
        new_x--;
    }
    if (direction == 3) {
        new_x++;
    }

    while(snk_led->child != NULL) {
        if (new_x == snk_led->x && new_y == snk_led->y) {
            return 0;
        }
        snk_led = snk_led->child;
    }
    return 1;
}

int main() {

    //INIT
    init_adc();
    init_interrupt();
    init_buzzer();

    struct Food *food = (Food*) malloc(sizeof(struct Food));

    struct led *snk_led = (led*) malloc(sizeof(struct led));
    snk_led->x = 4;
    snk_led->y = 4;
    snk_led->direction = 0;
    snk_led->parent = NULL;
    snk_led->child = NULL;

    struct led *snk_led_1 = (led*) malloc(sizeof(struct led));
    snk_led_1->x = 4;
    snk_led_1->y = 5;
    snk_led_1->direction = 0;
    snk_led_1->parent = snk_led;
    snk_led_1->parent->child = snk_led_1;
    snk_led_1->child = NULL;

    struct led *snk_led_2 = (led*) malloc(sizeof(struct led));
    snk_led_2->x = 4;
    snk_led_2->y = 6;
    snk_led_2->direction = 0;
    snk_led_2->parent = snk_led_1;
    snk_led_2->parent->child = snk_led_2;
    snk_led_2->child = NULL;

    DDRB |= (1 << PB0);
    DDRB |= (1 << PB1);
    DDRB |= (1 << PB3);

    int i;

    struct adc_res *adc_val = (adc_res*) malloc(sizeof(struct adc_res));
    generate_food(food, snk_led);

    for (i = 0; i < 5; i++)
        start_game_animation();
    /*while(x < 800) {
        start_game_animation();
    }*/

    /*while(PINB & (1 << PB2)) {
        start_game_animation();
    }*/


    init_7_segm();
    while(1) {
        if (isDead(snk_led) == 1) {
            deadAnimation();
        } else 
        {
            cli();
            ADCSRA |= (1 << ADEN);
            uint16_t value_x = x, value_y = y;
            ADCSRA &= ~(1 << ADEN);
            sei();
            //uint16_t value_x = 512, value_y = 512;

            /*if (snk_led->y > food->y) {
                value_y = 900;
            }
            if (snk_led->y < food->y) {
                value_y = 100;
            }
            if (snk_led->y == food->y && snk_led->x > food_x) {
                value_x = 100;
            }
            if (snk_led->y == food->y && snk_led->x < food_x) {
                value_x = 900;
            }*/

            int cont = 0, new_dir = snk_led->direction;

            if (value_x < 200) {
                if (snk_led->direction != 2)
                    new_dir = 3;
            } else if (value_x > 800){
                if (snk_led->direction != 3)
                    new_dir = 2;
            }

            if (value_y < 200) {
                if (snk_led->direction != 1)
                    new_dir = 0;
            } else if (value_y > 800){
                if (snk_led->direction != 0)
                    new_dir = 1;
            }

            /*while (1) {
                cont++;
                if (is_move_right(snk_led, new_dir) == 1) {
                    break;
                } else {
                    new_dir = (new_dir + 1) % 4;
                }
                //if (cont == 4)
                  //  break;
            }*/
            snk_led->direction = new_dir;
            move_snake(snk_led, adc_val, food);
        }
    }

    return 0;
}